"""
This module contains code modified from the CascadePSP project
The implementation is based on the segmentation-refinement code
Original code: https://github.com/hkchengrex/CascadePSP
"""

import torch
import torch.nn.functional as F

def resize_max_side(im, size, method):
    h, w = im.shape[-2:]
    max_side = max(h, w)
    ratio = size / max_side
    if method in ['bilinear', 'bicubic']:
        return F.interpolate(im, scale_factor=ratio, mode=method, align_corners=False)
    else:
        return F.interpolate(im, scale_factor=ratio, mode=method)

def safe_forward(model, im, seg, inter_s8=None, inter_s4=None):
    """
    Slightly pads the input image such that its length is a multiple of 8
    """
    b, _, ph, pw = seg.shape
    if (ph % 8 != 0) or (pw % 8 != 0):
        newH = ((ph//8+1)*8)
        newW = ((pw//8+1)*8)
        p_im = torch.zeros(b, 3, newH, newW, device=im.device)
        p_seg = torch.zeros(b, 1, newH, newW, device=im.device) - 1

        p_im[:,:,0:ph,0:pw] = im
        p_seg[:,:,0:ph,0:pw] = seg
        im = p_im
        seg = p_seg

        if inter_s8 is not None:
            p_inter_s8 = torch.zeros(b, 1, newH, newW, device=im.device) - 1
            p_inter_s8[:,:,0:ph,0:pw] = inter_s8
            inter_s8 = p_inter_s8
        if inter_s4 is not None:
            p_inter_s4 = torch.zeros(b, 1, newH, newW, device=im.device) - 1
            p_inter_s4[:,:,0:ph,0:pw] = inter_s4
            inter_s4 = p_inter_s4

    images = model(im, seg, inter_s8, inter_s4)
    return_im = {}

    for key in ['pred_224', 'pred_28_3', 'pred_56_2']:
        return_im[key] = images[key][:,:,0:ph,0:pw]
    del images

    return return_im

def process_high_res_im(model, im, seg, L=900):

    stride = L//2

    _, _, h, w = seg.shape

    """
    Global Step
    """
    if max(h, w) > L:
        im_small = resize_max_side(im, L, 'area')
        seg_small = resize_max_side(seg, L, 'area')
    elif max(h, w) < L:
        im_small = resize_max_side(im, L, 'bicubic')
        seg_small = resize_max_side(seg, L, 'bilinear')
    else:
        im_small = im
        seg_small = seg

    images = safe_forward(model, im_small, seg_small)

    pred_224 = images['pred_224']
    pred_56 = images['pred_56_2']
    
    """
    Local step
    """

    for new_size in [max(h, w)]:
        im_small = resize_max_side(im, new_size, 'area')
        seg_small = resize_max_side(seg, new_size, 'area')
        _, _, h, w = seg_small.shape

        combined_224 = torch.zeros_like(seg_small)
        combined_weight = torch.zeros_like(seg_small)

        r_pred_224 = (F.interpolate(pred_224, size=(h, w), mode='bilinear', align_corners=False)>0.5).float()*2-1
        r_pred_56 = F.interpolate(pred_56, size=(h, w), mode='bilinear', align_corners=False)*2-1

        padding = 16
        step_size = stride - padding*2
        step_len  = L

        used_start_idx = {}
        for x_idx in range((w)//step_size+1):
            for y_idx in range((h)//step_size+1):

                start_x = x_idx * step_size
                start_y = y_idx * step_size
                end_x = start_x + step_len
                end_y = start_y + step_len

                # Shift when required
                if end_y > h:
                    end_y = h
                    start_y = h - step_len
                if end_x > w:
                    end_x = w
                    start_x = w - step_len

                # Bound x/y range
                start_x = max(0, start_x)
                start_y = max(0, start_y)
                end_x = min(w, end_x)
                end_y = min(h, end_y)

                # The same crop might appear twice due to bounding/shifting
                start_idx = start_y*w + start_x
                if start_idx in used_start_idx:
                    continue
                else:
                    used_start_idx[start_idx] = True
                
                # Take crop
                im_part = im_small[:,:,start_y:end_y, start_x:end_x]
                seg_224_part = r_pred_224[:,:,start_y:end_y, start_x:end_x]
                seg_56_part = r_pred_56[:,:,start_y:end_y, start_x:end_x]

                # Skip when it is not an interesting crop anyway
                seg_part_norm = (seg_224_part>0).float()
                high_thres = 0.9
                low_thres = 0.1
                if (seg_part_norm.mean() > high_thres) or (seg_part_norm.mean() < low_thres):
                    continue
                grid_images = safe_forward(model, im_part, seg_224_part, seg_56_part)
                grid_pred_224 = grid_images['pred_224']

                # Padding
                pred_sx = pred_sy = 0
                pred_ex = step_len
                pred_ey = step_len

                if start_x != 0:
                    start_x += padding
                    pred_sx += padding
                if start_y != 0:
                    start_y += padding
                    pred_sy += padding
                if end_x != w:
                    end_x -= padding
                    pred_ex -= padding
                if end_y != h:
                    end_y -= padding
                    pred_ey -= padding

                combined_224[:,:,start_y:end_y, start_x:end_x] += grid_pred_224[:,:,pred_sy:pred_ey,pred_sx:pred_ex]

                del grid_pred_224

                # Used for averaging
                combined_weight[:,:,start_y:end_y, start_x:end_x] += 1

        # Final full resolution output
        seg_norm = (r_pred_224/2+0.5)
        pred_224 = combined_224 / combined_weight
        pred_224 = torch.where(combined_weight==0, seg_norm, pred_224)

    _, _, h, w = seg.shape
    images = {}
    images['pred_224'] = F.interpolate(pred_224, size=(h, w), mode='bilinear', align_corners=True)

    return images['pred_224']


def process_im_single_pass(model, im, seg, L=900):
    """
    A single pass version, aka global step only.
    """

    _, _, h, w = im.shape
    if max(h, w) < L:
        im = resize_max_side(im, L, 'bicubic')
        seg = resize_max_side(seg, L, 'bilinear')

    if max(h, w) > L:
        im = resize_max_side(im, L, 'area')
        seg = resize_max_side(seg, L, 'area')

    images = safe_forward(model, im, seg)

    if max(h, w) < L:
        images['pred_224'] = F.interpolate(images['pred_224'], size=(h, w), mode='area')
    elif max(h, w) > L:
        images['pred_224'] = F.interpolate(images['pred_224'], size=(h, w), mode='bilinear', align_corners=True)

    return images['pred_224']



import click
import sys
import os

from . import __version__


@click.command(context_settings={"help_option_names": ["-h", "--help"]})
@click.version_option(__version__, "-v", "--version", prog_name="clipx")
@click.option("-i", "--input", required=True, help="Path to input image")
@click.option("-o", "--output", required=False,
              help="Path to output image (defaults to input filename + '_remove.png' in the same directory)")
@click.option("--mask", is_flag=True, help="Output only mask")
@click.option("--no-refine", is_flag=True, help="Skip mask refinement")
@click.option("--device", default=None, help="Device to use (e.g., 'cpu', 'cuda:0'), defaults to 'cuda:0' if available")
@click.option("--fast", is_flag=True, help="Use fast mode for refinement")
def main(input, output, mask, no_refine, device, fast):
    """
    CLI tool for removing background from images
    """
    if not os.path.isfile(input):
        sys.exit(f"Error: Input file '{input}' does not exist")

    if output is None:
        # Use current working directory instead of input directory
        input_filename = os.path.basename(input)
        base_name = os.path.splitext(input_filename)[0]
        output = os.path.join(os.getcwd(), f"{base_name}_remove.png")

    output_dir = os.path.dirname(output)
    if output_dir and not os.path.exists(output_dir):
        try:
            os.makedirs(output_dir, exist_ok=True)
        except Exception as e:
            sys.exit(f"Error: Cannot create output directory: {e}")

    try:
        # Only import remove function when actually processing an image
        from .core import remove

        result = remove(
            input,
            only_mask=mask,
            refine_mask=not no_refine,
            device=device,
            fast=fast,
            L=900,
        )

        result.save(output)
    except Exception as e:
        sys.exit(f"Error processing image: {e}")


"""
This module contains code modified from the CascadePSP project
The implementation is based on the segmentation-refinement code
Original code: https://github.com/hkchengrex/CascadePSP
"""

from .main import Refiner


__version__ = '0.4.2'

class LazyLoader:
    def __init__(self, lib_name):
        self.lib_name = lib_name
        self._lib = None

    def __getattr__(self, name):
        if self._lib is None:
            import importlib
            self._lib = importlib.import_module(self.lib_name)
        return getattr(self._lib, name)


core = LazyLoader('.core')


def remove(*args, **kwargs):
    return core.remove(*args, **kwargs)


"""
This module contains code modified from the CascadePSP project
The implementation is based on the segmentation-refinement code
Original code: https://github.com/hkchengrex/CascadePSP
"""

import os
import requests
import hashlib


def download_model(destination):
    print('Downloading model file into', destination)
    URL = 'https://github.com/zhangsdong/clipx/releases/download/v0.1.0/model'

    with open(destination, 'wb') as file:
        response = requests.get(URL)
        file.write(response.content)

    print('Download completed.')


def check_model(destination):
    correct_md5 = '705806e3f26b039ee17c7f5b851c33ad'

    with open(destination, 'rb') as f:
        md5_returned = hashlib.md5(f.read()).hexdigest()

    return correct_md5 == md5_returned


def download_and_or_check_model_file(destination):

    if os.path.exists(destination):
        # print('Model file exists.')
        pass
    else:
        print('Model does not exist.')
        download_model(destination)

    md5_passed = check_model(destination)

    if md5_passed:
        return

    print('MD5 of the model file does not match')
    print('Downloading the model again...')
    download_model(destination)

    md5_passed = check_model(destination)
    assert md5_passed, 'MD5 still does not pass'
    




import os
import numpy as np
from PIL import Image
import onnxruntime as ort
from typing import Tuple, List, Optional

from .download import download_and_or_check_model_file


class U2NetPredictor:
    def __init__(self, model_path: str = None, providers: List[str] = None,
                 download_and_check_model: bool = True):
        """Initialize model session"""
        # Default model path if not specified
        if model_path is None:
            model_folder = os.path.expanduser("~/.clipx/u2net")
            if not os.path.exists(model_folder):
                os.makedirs(model_folder, exist_ok=True)
            model_path = os.path.join(model_folder, 'u2net.onnx')

        # Download model if needed
        if download_and_check_model:
            download_and_or_check_model_file(model_path)

        # Check if model exists
        if not os.path.exists(model_path):
            raise FileNotFoundError(f"Model file not found: {model_path}")

        # Set default providers if not specified
        if providers is None:
            providers = ['CPUExecutionProvider']

        # Load model
        self.session = self._load_model(model_path, providers)

    def _load_model(self, model_path: str, providers: List[str]):
        """Load ONNX model"""
        sess_opts = ort.SessionOptions()
        return ort.InferenceSession(model_path, sess_options=sess_opts, providers=providers)

    def _normalize_image(self, img: Image.Image, mean: Tuple[float, float, float], std: Tuple[float, float, float],
                         size: Tuple[int, int]) -> dict:
        """Normalize and resize input image"""
        im = img.convert("RGB").resize(size, Image.Resampling.LANCZOS)
        im_array = np.array(im) / 255.0
        im_array = (im_array - mean) / std
        im_array = im_array.transpose((2, 0, 1)).astype(np.float32)
        return {self.session.get_inputs()[0].name: np.expand_dims(im_array, 0)}

    def predict_mask(self, img: Image.Image) -> Image.Image:
        """Generate background mask"""
        input_tensor = self._normalize_image(img, (0.485, 0.456, 0.406), (0.229, 0.224, 0.225), (320, 320))
        pred = self.session.run(None, input_tensor)[0][0, 0]
        pred = (pred - pred.min()) / (pred.max() - pred.min())
        mask = Image.fromarray((pred * 255).astype("uint8"), mode="L")
        return mask.resize(img.size, Image.Resampling.LANCZOS)

    def remove_background(self, img: Image.Image,
                          bgcolor: Optional[Tuple[int, int, int, int]] = (0, 0, 0, 0)) -> Image.Image:
        """Remove background and return image with transparent background"""
        mask = self.predict_mask(img)
        cutout = Image.composite(img.convert("RGBA"), Image.new("RGBA", img.size, bgcolor), mask)
        return cutout


"""
This module contains code modified from the CascadePSP project
The implementation is based on the segmentation-refinement code
Original code: https://github.com/hkchengrex/CascadePSP
"""

import torch
from torch import nn
from torch.nn import functional as F

from segmentation_refinement.models.psp import extractors


class PSPModule(nn.Module):
    def __init__(self, features, out_features=1024, sizes=(1, 2, 3, 6)):
        super().__init__()
        self.stages = []
        self.stages = nn.ModuleList([self._make_stage(features, size) for size in sizes])
        self.bottleneck = nn.Conv2d(features * (len(sizes) + 1), out_features, kernel_size=1)
        self.relu = nn.ReLU(inplace=True)

    def _make_stage(self, features, size):
        prior = nn.AdaptiveAvgPool2d(output_size=(size, size))
        conv = nn.Conv2d(features, features, kernel_size=1, bias=False)
        return nn.Sequential(prior, conv)

    def forward(self, feats):
        h, w = feats.size(2), feats.size(3)
        set_priors = [F.interpolate(input=stage(feats), size=(h, w), mode='bilinear', align_corners=False) for stage in self.stages]
        priors = set_priors + [feats]
        bottle = self.bottleneck(torch.cat(priors, 1))
        return self.relu(bottle)


class PSPUpsample(nn.Module):
    def __init__(self, x_channels, in_channels, out_channels):
        super().__init__()
        self.conv = nn.Sequential(
            nn.BatchNorm2d(in_channels),
            nn.ReLU(inplace=True),
            nn.Conv2d(in_channels, out_channels, 3, padding=1),
            nn.BatchNorm2d(out_channels),
            nn.ReLU(inplace=True),
            nn.Conv2d(out_channels, out_channels, 3, padding=1),
        )

        self.conv2 = nn.Sequential(
            nn.BatchNorm2d(out_channels),
            nn.ReLU(inplace=True),
            nn.Conv2d(out_channels, out_channels, 3, padding=1),
            nn.BatchNorm2d(out_channels),
            nn.ReLU(inplace=True),
            nn.Conv2d(out_channels, out_channels, 3, padding=1),
        )

        self.shortcut = nn.Conv2d(x_channels, out_channels, kernel_size=1)

    def forward(self, x, up):
        x = F.interpolate(input=x, scale_factor=2, mode='bilinear', align_corners=False)

        p = self.conv(torch.cat([x, up], 1))
        sc = self.shortcut(x)

        p = p + sc

        p2 = self.conv2(p)

        return p + p2


class RefinementModule(nn.Module):
    def __init__(self):
        super().__init__()

        self.feats = extractors.resnet50()
        self.psp = PSPModule(2048, 1024, (1, 2, 3, 6))

        self.up_1 = PSPUpsample(1024, 1024+256, 512)
        self.up_2 = PSPUpsample(512, 512+64, 256)
        self.up_3 = PSPUpsample(256, 256+3, 32)
        
        self.final_28 = nn.Sequential(
            nn.Conv2d(1024, 32, kernel_size=1),
            nn.ReLU(inplace=True),
            nn.Conv2d(32, 1, kernel_size=1),
        )

        self.final_56 = nn.Sequential(
            nn.Conv2d(512, 32, kernel_size=1),
            nn.ReLU(inplace=True),
            nn.Conv2d(32, 1, kernel_size=1),
        )

        self.final_11 = nn.Conv2d(32+3, 32, kernel_size=1)
        self.final_21 = nn.Conv2d(32, 1, kernel_size=1)

    def forward(self, x, seg, inter_s8=None, inter_s4=None):

        images = {}

        """
        First iteration, s8 output
        """
        if inter_s8 is None:
            p = torch.cat((x, seg, seg, seg), 1)

            f, f_1, f_2 = self.feats(p) 
            p = self.psp(f)

            inter_s8 = self.final_28(p)
            r_inter_s8 = F.interpolate(inter_s8, scale_factor=8, mode='bilinear', align_corners=False)
            r_inter_tanh_s8 = torch.tanh(r_inter_s8)

            images['pred_28'] = torch.sigmoid(r_inter_s8)
            images['out_28'] = r_inter_s8
        else:
            r_inter_tanh_s8 = inter_s8

        """
        Second iteration, s8 output
        """
        if inter_s4 is None:
            p = torch.cat((x, seg, r_inter_tanh_s8, r_inter_tanh_s8), 1)

            f, f_1, f_2 = self.feats(p) 
            p = self.psp(f)
            inter_s8_2 = self.final_28(p)
            r_inter_s8_2 = F.interpolate(inter_s8_2, scale_factor=8, mode='bilinear', align_corners=False)
            r_inter_tanh_s8_2 = torch.tanh(r_inter_s8_2)

            p = self.up_1(p, f_2)

            inter_s4 = self.final_56(p)
            r_inter_s4 = F.interpolate(inter_s4, scale_factor=4, mode='bilinear', align_corners=False)
            r_inter_tanh_s4 = torch.tanh(r_inter_s4)

            images['pred_28_2'] = torch.sigmoid(r_inter_s8_2)
            images['out_28_2'] = r_inter_s8_2
            images['pred_56'] = torch.sigmoid(r_inter_s4)
            images['out_56'] = r_inter_s4
        else:
            r_inter_tanh_s8_2 = inter_s8
            r_inter_tanh_s4 = inter_s4

        """
        Third iteration, s1 output
        """
        p = torch.cat((x, seg, r_inter_tanh_s8_2, r_inter_tanh_s4), 1)

        f, f_1, f_2 = self.feats(p) 
        p = self.psp(f)
        inter_s8_3 = self.final_28(p)
        r_inter_s8_3 = F.interpolate(inter_s8_3, scale_factor=8, mode='bilinear', align_corners=False)

        p = self.up_1(p, f_2)
        inter_s4_2 = self.final_56(p)
        r_inter_s4_2 = F.interpolate(inter_s4_2, scale_factor=4, mode='bilinear', align_corners=False)
        p = self.up_2(p, f_1)
        p = self.up_3(p, x)


        """
        Final output
        """
        p = F.relu(self.final_11(torch.cat([p, x], 1)), inplace=True)
        p = self.final_21(p)

        pred_224 = torch.sigmoid(p)

        images['pred_224'] = pred_224
        images['out_224'] = p
        images['pred_28_3'] = torch.sigmoid(r_inter_s8_3)
        images['pred_56_2'] = torch.sigmoid(r_inter_s4_2)
        images['out_28_3'] = r_inter_s8_3
        images['out_56_2'] = r_inter_s4_2

        return images



"""
This module contains code modified from the CascadePSP project
The implementation is based on the segmentation-refinement code
Original code: https://github.com/hkchengrex/CascadePSP
"""

import cv2
import time
import matplotlib.pyplot as plt
import cascadepsp as refine
image = cv2.imread('test/aeroplane.jpg')
mask = cv2.imread('test/aeroplane.png', cv2.IMREAD_GRAYSCALE)

# model_path can also be specified here
# This step takes some time to load the model
refiner = refine.Refiner(device='cuda:0') # device can also be 'cpu'

# Fast - Global step only.
# Smaller L -> Less memory usage; faster in fast mode.
mask_output = refiner.refine(image, mask, fast=False, L=900)

plt.imshow(mask_output)
plt.show()


# Core functionality and utilities
import numpy as np
import cv2
import torch
from PIL import Image
from typing import Union, Optional
from functools import lru_cache

# Global variables for model caching
_u2net_model = None
_refiner_models = {}


# Model management functions
def get_u2net_model():
    """Returns singleton instance of U2NetPredictor"""
    global _u2net_model
    if _u2net_model is None:
        # Load model only when needed
        from .models.u2net.main import U2NetPredictor
        _u2net_model = U2NetPredictor()
    return _u2net_model


def get_refiner_model(device: str):
    """Returns singleton instance of Refiner for specific device"""
    global _refiner_models
    if device not in _refiner_models:
        # Load model only when needed
        from .models.cascadepsp.main import Refiner
        _refiner_models[device] = Refiner(device=device)
    return _refiner_models[device]


def get_default_device() -> str:
    """Returns 'cuda:0' if available, else 'cpu'"""
    if torch.cuda.is_available():
        return "cuda:0"
    return "cpu"


def preprocess_image(image: Union[Image.Image, str]) -> tuple:
    """Preprocesses the input image and returns normalized arrays"""
    # Handle image input
    if isinstance(image, str):
        image = Image.open(image)

    # Convert to numpy arrays for processing
    img_rgb = np.array(image.convert("RGB"))

    return image, img_rgb


def create_transparent_output(img_original, refined_mask):
    """Creates transparent output image from original image and refined mask"""
    # Convert mask to binary format
    refined_mask_binary = (refined_mask > 127).astype(np.uint8)

    # Handle different image formats
    img_original_np = np.array(img_original)
    if len(img_original_np.shape) == 2:  # Grayscale
        img_cv = cv2.cvtColor(img_original_np, cv2.COLOR_GRAY2BGR)
    elif img_original_np.shape[2] == 4:  # RGBA
        img_cv = cv2.cvtColor(img_original_np, cv2.COLOR_RGBA2BGR)
    elif img_original_np.shape[2] == 3:  # RGB
        img_cv = cv2.cvtColor(img_original_np, cv2.COLOR_RGB2BGR)
    else:
        # Fallback for other formats
        img_cv = cv2.cvtColor(np.array(img_original.convert("RGB")), cv2.COLOR_RGB2BGR)

    # Apply mask to extract foreground
    foreground = cv2.bitwise_and(img_cv, img_cv, mask=refined_mask_binary)

    # Add alpha channel
    result_with_alpha = np.dstack((foreground, refined_mask_binary * 255))

    # Convert back to PIL Image with RGBA format
    result = Image.fromarray(cv2.cvtColor(result_with_alpha, cv2.COLOR_BGRA2RGBA))

    return result


def remove(
        image: Union[Image.Image, str],
        only_mask: bool = False,
        refine_mask: bool = True,
        device: Optional[str] = None,
        fast: bool = False,
        L: int = 900,
) -> Image.Image:
    """
    Removes background from an image

    Args:
        image: PIL Image or path to image file
        only_mask: If True, returns only the mask
        refine_mask: If True, refines the mask using CascadePSP
        device: Device to use for CascadePSP ('cpu' or 'cuda:0')
        fast: If True, uses fast mode for CascadePSP
        L: Hyperparameter for CascadePSP

    Returns:
        PIL Image with transparent background or mask
    """
    # Preprocess image
    original_image, img_rgb = preprocess_image(image)

    # Step 1: Get initial mask using U2Net (singleton model)
    predictor = get_u2net_model()
    mask = predictor.predict_mask(original_image)
    mask_np = np.array(mask.convert("L"))

    # Step 2: Refine mask if requested
    if refine_mask:
        # Use default device if none specified
        if device is None:
            device = get_default_device()

        # Get singleton refiner model for specific device
        refiner = get_refiner_model(device)
        refined_mask = refiner.refine(img_rgb, mask_np, fast=fast, L=L)
    else:
        refined_mask = mask_np

    # Return only the mask if requested
    if only_mask:
        return Image.fromarray(refined_mask)

    # Step 3: Create transparent image
    return create_transparent_output(original_image, refined_mask)


"""
This module contains code modified from the CascadePSP project
The implementation is based on the segmentation-refinement code
Original code: https://github.com/hkchengrex/CascadePSP
"""

import os

import numpy as np
import torch
from torchvision import transforms

from .models.psp.pspnet import RefinementModule
from .eval_helper import process_high_res_im, process_im_single_pass
from .download import download_and_or_check_model_file


class Refiner:
    def __init__(self, device='cpu', model_folder=None, download_and_check_model=True):
        """
        Initialize the segmentation refinement model.
        device can be 'cpu' or 'cuda'
        model_folder specifies the folder in which the model will be downloaded and stored. Defaulted in ~/.segmentation-refinement.
        """
        self.model = RefinementModule()
        self.device = device
        if model_folder is None:
            model_folder = os.path.expanduser("~/.clipx/cascadepsp")

        if not os.path.exists(model_folder):
            os.makedirs(model_folder, exist_ok=True)

        model_path = os.path.join(model_folder, 'model')
        if download_and_check_model:
            download_and_or_check_model_file(model_path)

        model_dict = torch.load(model_path, map_location={'cuda:0': device})
        new_dict = {}
        for k, v in model_dict.items():
            name = k[7:] # Remove module. from dataparallel
            new_dict[name] = v
        self.model.load_state_dict(new_dict)
        self.model.eval().to(device)

        self.im_transform = transforms.Compose([
            transforms.ToTensor(),
            transforms.Normalize(
                mean=[0.485, 0.456, 0.406],
                std=[0.229, 0.224, 0.225]
            ),
        ])

        self.seg_transform = transforms.Compose([
            transforms.ToTensor(),
            transforms.Normalize(
                mean=[0.5],
                std=[0.5]
            ),
        ])

    def refine(self, image, mask, fast=False, L=900):
        with torch.no_grad():
            """
            Refines an input segmentation mask of the image.

            image should be of size [H, W, 3]. Range 0~255.
            Mask should be of size [H, W] or [H, W, 1]. Range 0~255. We will make the mask binary by thresholding at 127.
            Fast mode - Use the global step only. Default: False. The speedup is more significant for high resolution images.
            L - Hyperparameter. Setting a lower value reduces memory usage. In fast mode, a lower L will make it runs faster as well.
            """
            image = self.im_transform(image).unsqueeze(0).to(self.device)
            mask = self.seg_transform((mask>127).astype(np.uint8)*255).unsqueeze(0).to(self.device)
            if len(mask.shape) < 4:
                mask = mask.unsqueeze(0)

            if fast:
                output = process_im_single_pass(self.model, image, mask, L)
            else:
                output = process_high_res_im(self.model, image, mask, L)

            return (output[0,0].cpu().numpy()*255).astype('uint8')



"""
This module contains code modified from the CascadePSP project
The implementation is based on the segmentation-refinement code
Original code: https://github.com/hkchengrex/CascadePSP
"""

from collections import OrderedDict
import math

import torch
import torch.nn as nn
import torch.nn.functional as F


def conv3x3(in_planes, out_planes, stride=1, dilation=1):
    return nn.Conv2d(in_planes, out_planes, kernel_size=3, stride=stride,
                     padding=dilation, dilation=dilation, bias=False)


class Bottleneck(nn.Module):
    expansion = 4

    def __init__(self, inplanes, planes, stride=1, downsample=None, dilation=1):
        super(Bottleneck, self).__init__()
        self.conv1 = nn.Conv2d(inplanes, planes, kernel_size=1, bias=False)
        self.bn1 = nn.BatchNorm2d(planes)
        self.conv2 = nn.Conv2d(planes, planes, kernel_size=3, stride=stride, dilation=dilation,
                               padding=dilation, bias=False)
        self.bn2 = nn.BatchNorm2d(planes)
        self.conv3 = nn.Conv2d(planes, planes * 4, kernel_size=1, bias=False)
        self.bn3 = nn.BatchNorm2d(planes * 4)
        self.relu = nn.ReLU(inplace=True)
        self.downsample = downsample
        self.stride = stride

    def forward(self, x):
        residual = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)
        out = self.relu(out)

        out = self.conv3(out)
        out = self.bn3(out)

        if self.downsample is not None:
            residual = self.downsample(x)

        out += residual
        out = self.relu(out)

        return out


class ResNet(nn.Module):
    def __init__(self, block, layers=(3, 4, 23, 3)):
        self.inplanes = 64
        super(ResNet, self).__init__()
        self.conv1 = nn.Conv2d(6, 64, kernel_size=7, stride=2, padding=3,
                               bias=False)
        self.bn1 = nn.BatchNorm2d(64)
        self.relu = nn.ReLU(inplace=True)
        self.maxpool = nn.MaxPool2d(kernel_size=3, stride=2, padding=1)
        self.layer1 = self._make_layer(block, 64, layers[0])
        self.layer2 = self._make_layer(block, 128, layers[1], stride=2)
        self.layer3 = self._make_layer(block, 256, layers[2], stride=1, dilation=2)
        self.layer4 = self._make_layer(block, 512, layers[3], stride=1, dilation=4)

        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                n = m.kernel_size[0] * m.kernel_size[1] * m.out_channels
                m.weight.data.normal_(0, math.sqrt(2. / n))
            elif isinstance(m, nn.BatchNorm2d):
                m.weight.data.fill_(1)
                m.bias.data.zero_()

    def _make_layer(self, block, planes, blocks, stride=1, dilation=1):
        downsample = None
        if stride != 1 or self.inplanes != planes * block.expansion:
            downsample = nn.Sequential(
                nn.Conv2d(self.inplanes, planes * block.expansion,
                          kernel_size=1, stride=stride, bias=False),
                nn.BatchNorm2d(planes * block.expansion),
            )

        layers = [block(self.inplanes, planes, stride, downsample)]
        self.inplanes = planes * block.expansion
        for i in range(1, blocks):
            layers.append(block(self.inplanes, planes, dilation=dilation))

        return nn.Sequential(*layers)

    def forward(self, x):
        x_1 = self.conv1(x)  # /2
        x = self.bn1(x_1)
        x = self.relu(x)
        x = self.maxpool(x)  # /2

        x_2 = self.layer1(x)
        x = self.layer2(x_2)   # /2
        x = self.layer3(x)
        x = self.layer4(x)

        return x, x_1, x_2


def resnet50():
    model = ResNet(Bottleneck, [3, 4, 6, 3])
    return model

